import copy
import inspect
import logging

from ..utils.utils import setup_compute
from .seq_scheduler import LocalSequentialScheduler

logger = logging.getLogger(__name__)

schedulers = {
    "local": LocalSequentialScheduler,
}

_scheduler_presets = {
    "auto": {"scheduler": "local", "searcher": "local_random"},
    "local_random": {"scheduler": "local", "searcher": "local_random"},
    "random": {"scheduler": "local", "searcher": "random"},
}


def compile_scheduler_options(
    scheduler_options,
    nthreads_per_trial,
    ngpus_per_trial,
    num_trials,
    time_out,
    scheduler=None,
    search_strategy=None,
    search_options=None,
    checkpoint=None,
    resume=False,
    visualizer=None,
    time_attr=None,
    reward_attr=None,
    dist_ip_addrs=None,
    epochs=None,
):
    """
    Updates a copy of scheduler_options (scheduler-specific options, can be
    empty) with general options. The result can be passed to __init__ of the
    scheduler.

    Special role of epochs for HyperbandScheduler: If the search_strategy
    involves HyperbandScheduler and epochs is given, then this value is
    copied to scheduler_options['max_t']. Pass epochs for applications
    where the time_attr is epoch, and epochs is the maximum number of
    epochs.

    :param scheduler_options:
    :param scheduler:
    :param search_strategy:
    :param search_options:
    :param nthreads_per_trial:
    :param ngpus_per_trial:
    :param checkpoint:
    :param num_trials:
    :param time_out:
    :param resume:
    :param visualizer:
    :param time_attr:
    :param reward_attr:
    :param dist_ip_addrs:
    :param kwargs:
    :param epochs: See above. Optional
    :return: Copy of scheduler_options with updates

    """
    if scheduler_options is None:
        scheduler_options = dict()
    else:
        assert isinstance(scheduler_options, dict)
    scheduler_options = copy.copy(scheduler_options)
    if dist_ip_addrs is None:
        dist_ip_addrs = []
    if search_strategy is None:
        search_strategy = "random"
    if scheduler is None:
        scheduler = "local"
    assert isinstance(search_strategy, str)
    if search_options is None:
        search_options = dict()
    if visualizer is None:
        visualizer = "none"
    if time_attr is None:
        time_attr = "epoch"
    if reward_attr is None:
        reward_attr = "validation_performance"
    scheduler_params = {
        "resource": {"num_cpus": nthreads_per_trial, "num_gpus": ngpus_per_trial},
        "scheduler": scheduler,
        "searcher": search_strategy,
        "search_options": search_options,
        "checkpoint": checkpoint,
        "resume": resume,
        "num_trials": num_trials,
        "time_out": time_out,
        "reward_attr": reward_attr,
        "time_attr": time_attr,
        "visualizer": visualizer,
        "dist_ip_addrs": dist_ip_addrs,
    }
    resource = None
    if "resource" in scheduler_options:
        scheduler_params["resource"].update(scheduler_options["resource"])
        resource = scheduler_params["resource"].copy()
    scheduler_params.update(scheduler_options)
    if resource:
        scheduler_params["resource"] = resource

    scheduler_params["resource"]["num_cpus"], scheduler_params["resource"]["num_gpus"] = setup_compute(
        nthreads_per_trial=scheduler_params["resource"]["num_cpus"],
        ngpus_per_trial=scheduler_params["resource"]["num_gpus"],
    )  # TODO: use 'auto' downstream

    required_options = [
        "resource",
        "scheduler",
        "searcher",
        "search_options",
        "checkpoint",
        "resume",
        "num_trials",
        "time_out",
        "reward_attr",
        "time_attr",
        "visualizer",
        "dist_ip_addrs",
    ]
    missing_options = []
    for option in required_options:
        if option not in scheduler_params:
            missing_options.append(option)
    if missing_options:
        raise AssertionError(f"Missing required keys in scheduler_options: {missing_options}")
    return scheduler_params


def scheduler_factory(
    hyperparameter_tune_kwargs,
    time_out: float = None,
    num_trials: int = None,
    nthreads_per_trial="all",
    ngpus_per_trial="all",
    **kwargs,
):
    """
    Constructs a scheduler via lazy initialization based on the input hyperparameter_tune_kwargs.
    The output will contain the scheduler class and init arguments except for the `train_fn` argument, which must be specified downstream.

    Parameters
    ----------
    hyperparameter_tune_kwargs : str or dict
        Hyperparameter tuning strategy and kwargs.
        If None, then hyperparameter tuning will not be performed.
        Valid preset values:
            'auto': Uses the 'random' preset.
            'random': Performs HPO via random search using local scheduler.
        The 'searcher' key is required when providing a dict. Some schedulers may have different valid keys.
    time_out : float, default = None
        Same as hyperparameter_tune_kwargs['time_out']. Ignored if specified in hyperparameter_tune_kwargs.
        At least one of time_out or num_trials must be specified.
    num_trials : int, default = None
        Same as hyperparameter_tune_kwargs['time_out']. Ignored if specified in hyperparameter_tune_kwargs.
        At least one of time_out or num_trials must be specified.
    nthreads_per_trial : str or int, default = 'all'
        Same as hyperparameter_tune_kwargs['nthreads_per_trial']. Ignored if specified in hyperparameter_tune_kwargs.
        Number of CPU threads to use per HPO trial.
        Valid str values:
            'all': Use all CPU threads.
            'auto': Keep value as 'auto' in output, must be updated downstream.
        If None, use all CPU threads as in 'all'.
    ngpus_per_trial : str or int, default = 'all'
        Same as hyperparameter_tune_kwargs['ngpus_per_trial']. Ignored if specified in hyperparameter_tune_kwargs.
        Number of GPUs to use per HPO trial.
        Valid str values:
            'all': Use all GPUs.
            'auto': Keep value as 'auto' in output, must be updated downstream.
        If None, use 0 GPUs.
    **kwargs :
        Kwargs to specify any other scheduler parameters.
        A kwarg will be ignored if also specified in hyperparameter_tune_kwargs.

    Returns
    -------
    scheduler_cls : class :class:`autogluon.core.scheduler.TaskScheduler`, scheduler_params : dict
        scheduler_cls is the class of the scheduler that will be constructed.
        scheduler_params is the key word parameter arguments to pass to the Scheeduler class constructor when initializing a Scheduler object.
        To actually construct a Scheduler object, call `scheduler_cls(train_fn, **scheduler_params)`
        By default in scheduler_params: time_attr='epoch', reward_attr='validation_performance'
    """
    if hyperparameter_tune_kwargs is None:
        raise ValueError(f"hyperparameter_tune_kwargs cannot be None.")
    if isinstance(hyperparameter_tune_kwargs, str):
        hyperparameter_tune_kwargs = get_hyperparameter_tune_kwargs_preset(hyperparameter_tune_kwargs)
    if not isinstance(hyperparameter_tune_kwargs, dict):
        raise ValueError(
            f"hyperparameter_tune_kwargs must be of type str or dict, but is type: {type(hyperparameter_tune_kwargs)}"
        )
    if "scheduler" not in hyperparameter_tune_kwargs:
        raise ValueError(
            f"Required key 'scheduler' is not present in hyperparameter_tune_kwargs: {hyperparameter_tune_kwargs}"
        )
    if "searcher" not in hyperparameter_tune_kwargs:
        raise ValueError(
            f"Required key 'searcher' is not present in hyperparameter_tune_kwargs: {hyperparameter_tune_kwargs}"
        )
    if num_trials is None and time_out is not None:
        num_trials = 1000

    scheduler_params = compile_scheduler_options(
        scheduler_options=hyperparameter_tune_kwargs,
        nthreads_per_trial=nthreads_per_trial,
        ngpus_per_trial=ngpus_per_trial,
        num_trials=num_trials,
        time_out=time_out,
        **kwargs,
    )

    scheduler_cls = hyperparameter_tune_kwargs.get("scheduler", "unknown")
    if isinstance(scheduler_cls, str):
        scheduler_cls = get_scheduler_from_preset(scheduler_cls)
    if not inspect.isclass(scheduler_cls):
        raise ValueError(f"scheduler_cls must be a class, but was instead: {scheduler_cls}")

    if scheduler_params["time_out"] is None:
        scheduler_params.pop("time_out", None)
    return scheduler_cls, scheduler_params


def get_scheduler_from_preset(scheduler_cls):
    scheduler_cls = scheduler_cls.lower()
    if scheduler_cls not in schedulers.keys():
        raise ValueError(
            f"Required key 'scheduler' in hyperparameter_tune_kwargs must be one of the "
            f"values {schedulers.keys()}, but was instead: {scheduler_cls}"
        )
    scheduler_cls = schedulers.get(scheduler_cls)
    return scheduler_cls


def get_hyperparameter_tune_kwargs_preset(preset: str):
    # TODO: re-enable bayesopt after it's been implemented
    if preset == "bayesopt":
        logger.warning(
            f"Bayesopt hyperparameter tuning is currently disabled. Will use random hyperparameter tuning instead."
        )
        preset = "random"
    if preset not in _scheduler_presets:
        raise ValueError(
            f'Invalid hyperparameter_tune_kwargs preset value "{preset}". Valid presets: {list(_scheduler_presets.keys())}'
        )
    return _scheduler_presets[preset].copy()



from __future__ import annotations


class AbstractES:
    """
    Abstract early stopping class
    """

    def update(self, cur_round, is_best: bool = False) -> bool:
        raise NotImplementedError

    def early_stop(self, cur_round, is_best: bool = False) -> bool:
        raise NotImplementedError


class NoES(AbstractES):
    """
    Dummy early stopping method that never triggers early stopping
    """

    def update(self, cur_round: int, is_best: bool = False) -> bool:
        return self.early_stop(cur_round, is_best=is_best)

    def early_stop(self, cur_round: int, is_best: bool = False) -> bool:
        return False


class SimpleES(AbstractES):
    """
    Implements early stopping with fixed patience

    Parameters
    ----------
    patience : int, default 10
        If no improvement occurs in `patience` rounds or greater, self.early_stop will return True.
    """

    def __init__(self, patience: int = 10):
        self.patience = patience
        self.best_round = 0

    def update(self, cur_round: int, is_best: bool = False) -> bool:
        if is_best:
            self.best_round = cur_round
        return self.early_stop(cur_round, is_best=is_best)

    def early_stop(self, cur_round: int, is_best: bool = False) -> bool:
        if is_best:
            return False
        return cur_round - self.best_round >= self.patience


# TODO: Add time component
#  if given a large amount of time and training is fast, should check more rounds before early stopping
# TODO: Incorporate score, rolling window
class AdaptiveES(AbstractES):
    """
    Implements early stopping with adaptive patience

    Patience follows the formula `patience = ax + b`, where `a = adaptive_rate`, `x = round` and `b = adaptive_offset`.
    Patience is only updated when a new `best_round` is observed.

    Patience is adaptively adjusted across training instead of being a fixed value.
    This generally outperforms fixed patience strategies. Examples below:
    1. If the current best_round is 10000, it is reasonable to assume that it could take more than 100 rounds before finding a new best.
    2. If the current best_round is 3, it is unlikely that there will be 100 rounds before finding a new best at round 103.
    In the above examples, a fixed patience of 100 would be too little for round 10000, but too large for round 3.
    However, with `adaptive_rate=0.2`, `adaptive_offset=10`, round 3 would have a patience of ~10, while round 10000 would have a patience of ~2000.

    Parameters
    ----------
    adaptive_rate : float, default 0.3
        The rate of increase in patience.
        Set to 0 to disable, or negative to shrink patience during training.
    adaptive_offset : int, default 10
        The initial patience when cur_round is 0.
    min_patience : int | None, default None
        The minimum value of patience. Ignored if None.
    max_patience : int | None, default None
        The maximum value of patience. Ignored if None.

    Attributes
    ----------
    best_round : int
        The most recent round passed to self.update with `is_best=True`.
        Dictates patience and is used to determine if self.early_stop() returns True.
    patience : int
        If no improvement occurs in `patience` rounds or greater, self.early_stop will return True.
        patience is dictated by the following formula:
        patience = min(self.max_patience, (max(self.min_patience, round(self.best_round * self.adaptive_rate + self.adaptive_offset))))
        Effectively, patience = self.best_round * self.adaptive_rate + self.adaptive_offset, bound by min_patience and max_patience
    """

    def __init__(
        self,
        adaptive_rate: float = 0.3,
        adaptive_offset: int = 10,
        min_patience: int | None = None,
        max_patience: int | None = None,
    ):
        self.adaptive_rate = adaptive_rate
        self.adaptive_offset = adaptive_offset
        self.min_patience = min_patience
        self.max_patience = max_patience
        self.best_round = 0
        self.patience = self._update_patience(self.best_round)

    def update(self, cur_round: int, is_best: bool = False) -> bool:
        """
        Updates the state of the object. Identical to calling self.early_stop, but if `is_best=True`, it will set `self.best_round=cur_round`.
        If cur_round achieved a new best score, set `is_best=True`.
        Ideally, this should be called every round during training, with the output used to determine if the model should stop training.
        """
        if is_best:
            self.best_round = cur_round
            self.patience = self._update_patience(self.best_round)
        return self.early_stop(cur_round, is_best=is_best)

    def early_stop(self, cur_round: int, is_best: bool = False) -> bool:
        """
        Returns True if (cur_round - self.best_round) equals or exceeds self.patience, otherwise returns False.
        This can be used to indicate if training should stop.
        """
        if is_best:
            return False
        return cur_round - self.best_round >= self.patience

    def _update_patience(self, best_round: int) -> int:
        patience = round(self.adaptive_rate * best_round + self.adaptive_offset)  # ax + b
        if self.min_patience is not None:
            patience = max(self.min_patience, patience)
        if self.max_patience is not None:
            patience = min(self.max_patience, patience)
        return patience


ES_CLASS_MAP = {
    "simple": SimpleES,
    "adaptive": AdaptiveES,
}
